from dataclasses import dataclass
from typing import List, Optional

from ._char import Character
from ._repeat import InfiniteRepeat


@dataclass(frozen=True)
class Sequence:
    elements: List

    @property
    def starriness(self):
        return sum(e.starriness for e in self.elements)

    def __len__(self):
        return len(self.elements)

    def example(self) -> str:
        return "".join(e.example() for e in self.elements)

    @property
    def minimum_length(self) -> int:
        accum: int = 0
        for e in self.elements:
            accum += e.minimum_length
        return accum

    def exact_character_class(self) -> Optional[Character]:
        """
        aa*a -> a, abc -> None, [ab][abc] -> None
        """
        first = self.elements[0].exact_character_class()
        if first is None:
            return None
        for c in self.elements[1:]:
            if c != first:
                return None
        return c

    def overall_character_class(self) -> Optional[Character]:
        """
        aa*a -> a, abc -> None, [ab][abc] -> [ab]
        a?b -> b, a+b -> None, [ab]+b* -> b
        """
        c = Character.ANY()
        for e in self.elements:
            c &= e.overall_character_class()
            if not c:
                return None
        return c

    def matching_repeats(self):
        """Complicated way to get the possible character classes for a sequence"""
        c = Character.ANY()
        has_mandatory = False
        optionals = []
        starriness = 0
        minimum_length = 0
        for e in self.elements:
            if e.minimum_length:
                c &= e.overall_character_class()
                if not c:
                    return None
                has_mandatory = True
                starriness += e.starriness
                minimum_length += e.minimum_length
            elif e.starriness > 0:
                optionals.append(e)
        possibilities = {c: starriness} if has_mandatory else {}
        for e in optionals:
            if new_c := e.overall_character_class() & c:
                if new_c in possibilities:
                    possibilities[new_c] += e.starriness
                else:
                    possibilities[new_c] = e.starriness

        if len(possibilities) > 1:
            # (a*[ab]*a*[bc]*[bcd]*.+a*)*@ has classes {.: 1, [a]: 5, [[a-b]]: 2, [[b-c]]: 3, [[b-d]]: 2, [b]: 3}
            # This could blow up!
            poss_chars = list(possibilities.items())
            merged_chars = {}
            while poss_chars:
                c_a, s_a = poss_chars.pop()
                for c_b, s_b in poss_chars:
                    if (merged := c_a & c_b) is not None:
                        if merged == c_a:
                            possibilities[c_a] += s_b
                        elif merged == c_b:
                            possibilities[c_b] += s_a
                        else:
                            if merged not in merged_chars:
                                merged_chars[merged] = set()
                            merged_chars[merged] |= {(c_a, s_a), (c_b, s_b)}
            for merged, set_of_chars in merged_chars.items():
                possibilities[merged] = sum(s for _, s in set_of_chars)

        for cc, s in possibilities.items():
            if s:
                yield InfiniteRepeat(cc, minimum_length, forced_starriness=s)

    def maximal_character_class(self) -> Character:
        """
        Only useful when this Sequence is inside a Repeat
        a*b -> [ab], ab* -> [ab]
        Since forcing backtracking for (bc*)$
        """
        c = None
        for e in self.elements:
            if (mcc := e.maximal_character_class()) is not None:
                c = mcc | c
        return c

    def __repr__(self) -> str:
        return "SEQ{ " + " ".join(str(e) for e in self.elements) + " }"



from dataclasses import dataclass
from typing import Iterator, List, Optional

from ._at import EndOfString
from ._char import Character
from ._repeat import FiniteRepeat, InfiniteRepeat
from ._sequence import Sequence


@dataclass(frozen=True)
class Branch:
    branches: List
    optional: bool = False

    def get_branches(self) -> Iterator:
        for b in self.branches:
            yield b
        if self.optional:
            yield None

    @property
    def starriness(self) -> int:
        return max(b.starriness for b in self.branches)

    @property
    def minimum_length(self) -> int:
        return 0 if self.optional else min(b.minimum_length for b in self.branches)

    def overall_character_class(self) -> Optional[Character]:
        c = Character.ANY()
        for b in self.branches:
            c &= b.overall_character_class()
            if c is None:
                return None
        return c

    def maximal_character_class(self):
        return None  # Really?

    def example(self) -> str:
        if self.optional:
            return ""
        return self.branches[0].example()

    def __len__(self) -> int:
        return len(self.branches) + int(self.optional)

    def __repr__(self) -> str:
        middle = " | ".join(str(b) for b in self.branches)
        return f"BR( {middle} ){'?' if self.optional else ''}"

    def matching_repeats(self):
        for b in self.branches:
            if b.starriness > 0:
                if isinstance(b, InfiniteRepeat):
                    yield b
                elif isinstance(b, Sequence):
                    yield from b.matching_repeats()


def make_branch(branches: List):
    if len(branches) == 1:
        return branches[0]
    optional = False
    non_empty_branches = [b for b in branches if b and not isinstance(b, EndOfString)]
    if not non_empty_branches:
        return None
    if len(non_empty_branches) < len(branches):
        # (ab|cd|) -> (ab|cd)?
        optional = True
    if all(isinstance(b, Character) for b in non_empty_branches):
        # (a|b) -> [ab], (a|b|) -> [ab]?
        c = None
        for b in non_empty_branches:
            c |= b
        if optional:
            return FiniteRepeat(c, 0, 1)
        else:
            return c

    return Branch(non_empty_branches, optional)



"""PEP-517 compliant buildsystem API"""
import logging
import os
import shutil
import subprocess
import sysconfig
import tarfile
import tempfile
from gzip import GzipFile
from pathlib import Path

from wheel.wheelfile import WheelFile

from ._pyc_wheel import convert_wheel
from ._util import GET_CHECK
from ._util import _write_wheel_file
from ._util import cd
from ._util import install_files_path
from ._util import meson
from ._util import meson_configure
from .config import Config
from .metadata import get_python_bin
from .pep425tags import get_platform_tag

log = logging.getLogger(__name__)


def get_requires_for_build_wheel(config_settings=None):
    """Returns a list of requirements for building, as strings"""
    return Config().requirements


# For now, we require all dependencies to build either a wheel or an sdist.
get_requires_for_build_sdist = get_requires_for_build_wheel


def check_is_pure(installed):
    variables = sysconfig.get_config_vars()
    suffix = (
        variables.get('EXT_SUFFIX')
        or variables.get('SO')
        or variables.get('.so')
    )
    # msys2's python3 has "-cpython-36m.dll", we have to be clever
    split = suffix.rsplit('.', 1)
    suffix = split.pop(-1)

    for installpath in installed.values():
        if "site-packages" in installpath or "dist-packages" in installpath:
            if installpath.split('.')[-1] == suffix:
                return False

    return True


def prepare_metadata_for_build_wheel(
    metadata_directory, config_settings=None, builddir=None, config=None
):
    """Creates {metadata_directory}/foo-1.2.dist-info"""
    if not builddir:
        builddir = tempfile.TemporaryDirectory().name
        meson_configure(builddir, config_settings=config_settings)
    if not config:
        config = Config(builddir)

    dist_info = Path(
        metadata_directory,
        '{}-{}.dist-info'.format(config['module'].replace('-','_'), config['version']),
    )
    dist_info.mkdir(exist_ok=True)

    is_pure = check_is_pure(config.installed)
    with (dist_info / 'WHEEL').open('w') as f:
        _write_wheel_file(f, False, is_pure)

    with (dist_info / 'METADATA').open('w') as f:
        f.write(config.get_metadata())

    with (dist_info / config.license_file).open('w') as fw:
        with Path(config.license_file).open('r') as fr:
            fw.write(fr.read())

    if config.entry_points:
        with (dist_info / 'entry_points.txt').open('w') as f:
            f.write(config.entry_points)

    return dist_info.name


def get_abi(python):
    return (
        subprocess.check_output([python, '-c', GET_CHECK])
        .decode('utf-8')
        .strip('\n')
    )


class WheelBuilder:
    def __init__(self):
        self.wheel_zip = None  # type: ignore
        self.builddir = tempfile.TemporaryDirectory()
        self.installdir = tempfile.TemporaryDirectory()

    def build(self, wheel_directory, config_settings, metadata_dir):
        config = Config()

        args = [
            self.builddir.name,
            '--prefix',
            self.installdir.name,
        ] + config.get('meson-options', [])
        meson_configure(*args, config_settings=config_settings)
        config.builddir = self.builddir.name

        metadata_dir = prepare_metadata_for_build_wheel(
            wheel_directory, builddir=self.builddir.name, config=config
        )

        is_pure = check_is_pure(config.installed)
        platform_tag = config.get(
            'platforms', 'any' if is_pure else get_platform_tag()
        )
        python = get_python_bin(config)
        if not is_pure:
            abi = get_abi(python)
        else:
            abi = config.get('pure-python-abi', get_abi(python))
        target_fp = wheel_directory / '{}-{}-{}-{}.whl'.format(
            config['module'].replace('-','_'),
            config['version'],
            abi,
            platform_tag,
        )

        self.wheel_zip: WheelFile = WheelFile(str(target_fp), 'w')
        for f in os.listdir(str(wheel_directory / metadata_dir)):
            self.wheel_zip.write(
                str(wheel_directory / metadata_dir / f),
                arcname=str(Path(metadata_dir) / f),
            )
        shutil.rmtree(Path(wheel_directory) / metadata_dir)

        # Make sure everything is built
        meson('install', '-C', self.builddir.name)
        self.pack_files(config)
        self.wheel_zip.close()
        optimize, *_ = [i.get('value', -1) for i in config.options if i.get('name', '') == 'python.bytecompile']
        convert_wheel(Path(target_fp), optimize=optimize, **config.pyc_wheel)
        return str(target_fp)

    def pack_files(self, config):
        for _, installpath in config.installed.items():
            if "site-packages" in installpath:
                installpath = install_files_path(installpath, 'site-packages')
                self.wheel_zip.write_files(installpath)
                break
            elif "dist-packages" in installpath:
                installpath = install_files_path(installpath, 'dist-packages')
                self.wheel_zip.write_files(installpath)
                break


def build_wheel(
    wheel_directory, config_settings=None, metadata_directory=None
):
    """Builds a wheel, places it in wheel_directory"""
    return WheelBuilder().build(
        Path(wheel_directory), config_settings, metadata_directory
    )


def build_sdist(sdist_directory, config_settings=None):
    """Builds an sdist, places it in sdist_directory"""
    distdir = Path(sdist_directory)
    with tempfile.TemporaryDirectory() as builddir:
        with tempfile.TemporaryDirectory() as installdir:
            meson(
                builddir,
                '--prefix',
                installdir,
                config_settings=config_settings,
                builddir=builddir,
            )

            config = Config(builddir)
            meson('dist', '--no-tests', '-C', builddir)

            tf_dir = '{}-{}'.format(config['module'], config['version'])
            mesondistfilename = '%s.tar.xz' % tf_dir
            mesondisttar = tarfile.open(
                Path(builddir) / 'meson-dist' / mesondistfilename
            )
            for entry in mesondisttar:
                # GOOD: Check that entry is safe
                if os.path.isabs(entry.name) or ".." in entry.name:
                    raise ValueError("Illegal tar archive entry")
                mesondisttar.extract(entry, installdir)
            # OZI uses setuptools_scm to create PKG-INFO
            pkg_info = config.get_metadata()
            distfilename = '{}-{}.tar.gz'.format(config['module'], config['version'])
            target = distdir / distfilename
            source_date_epoch = os.environ.get('SOURCE_DATE_EPOCH', '')
            mtime = int(source_date_epoch) if source_date_epoch else None
            with GzipFile(str(target), mode='wb', mtime=mtime) as gz:
                with cd(installdir):
                    with tarfile.TarFile(
                        str(target),
                        mode='w',
                        fileobj=gz,
                        format=tarfile.PAX_FORMAT,
                    ) as tf:
                        tf.add(tf_dir, recursive=True)
                        pkginfo_path = Path(installdir) / tf_dir / 'PKG-INFO'
                        if not pkginfo_path.exists():
                            with open(pkginfo_path, mode='w') as fpkginfo:
                                fpkginfo.write(pkg_info)
                                fpkginfo.flush()
                                tf.add(Path(tf_dir) / 'PKG-INFO')
    return target.name



import json
import logging
import os
import sys

from ._util import check_pyproject_regexes
from .metadata import auto_python_version
from .metadata import check_pkg_info_file
from .metadata import check_requires_python
from .metadata import get_description_headers
from .metadata import get_download_url_headers
from .metadata import get_license_headers
from .metadata import get_optional_dependencies
from .metadata import get_python_bin
from .metadata import get_requirements_headers
from .metadata import get_simple_headers
from .schema import VALID_OPTIONS
from .schema import VALID_PYC_WHEEL_OPTIONS

if sys.version_info >= (3, 11):
    import tomllib as toml
elif sys.version_info < (3, 11):
    import tomli as toml

log = logging.getLogger(__name__)


class Config:
    def __init__(self, builddir=None):
        config = self.__get_config()
        check_pyproject_regexes(config)
        self.__metadata = config['tool']['ozi-build']['metadata']
        self.__entry_points = config['tool']['ozi-build'].get(
            'entry-points', []
        )
        self.__extras = config.get('project', {}).get('optional_dependencies', None)
        if self.__extras is not None:
            log.warning('pyproject.toml:project.optional_dependencies should be renamed to pyproject.toml:project.optional-dependencies')
        else:
            self.__extras = config.get('project', {}).get('optional-dependencies', {})
        self.__requires = config.get('project', {}).get('dependencies', None)
        self.license_file = config.get('project', {}).get('license', {}).get('file', '')
        if self.license_file == '':
            log.warning('pyproject.toml:project.license.file key-value pair was not found')
        self.__min_python = '3.10'
        self.__max_python = '3.13'
        self.__pyc_wheel = config['tool']['ozi-build'].get('pyc_wheel', {})
        self.installed = []
        self.options = []
        if builddir:
            self.builddir = builddir

    @property
    def extras(self):
        return self.__extras

    @property
    def min_python(self):
        return self.__min_python
    
    @property
    def max_python(self):
        return self.__max_python

    @property
    def requirements(self):
        return self.__requires if self.__requires else []

    @property
    def pyc_wheel(self):
        return self.__pyc_wheel

    def __introspect(self, introspect_type):
        with open(
            os.path.join(
                self.__builddir,
                'meson-info',
                'intro-' + introspect_type + '.json',
            )
        ) as f:
            return json.load(f)

    @staticmethod
    def __get_config():
        with open('pyproject.toml', 'rb') as f:
            config = toml.load(f)
            try:
                config['tool']['ozi-build']['metadata']
            except KeyError:
                raise RuntimeError(
                    "`[tool.ozi-build.metadata]` section is mandatory "
                    "for the meson backend"
                )

            return config

    def __getitem__(self, key):
        return self.__metadata[key]

    def __setitem__(self, key, value):
        self.__metadata[key] = value

    def __contains__(self, key):
        return key in self.__metadata

    @property
    def entry_points(self):
        res = ''
        for group_name in sorted(self.__entry_points):
            res += '[{}]\n'.format(group_name)
            group = self.__entry_points[group_name]
            for entrypoint in sorted(group):
                res += '{}\n'.format(entrypoint)
            res += '\n'
        return res

    @property
    def builddir(self):
        return self.__builddir

    @builddir.setter
    def builddir(self, builddir):
        self.__builddir = builddir
        project = self.__introspect('projectinfo')

        self['version'] = project['version']
        if 'module' not in self:
            self['module'] = project['descriptive_name']

        self.installed = self.__introspect('installed')
        self.options = self.__introspect('buildoptions')
        self.validate_options()

    def validate_options(self):
        options = VALID_OPTIONS.copy()
        options['version'] = {}
        options['module'] = {}
        for field, value in self.__metadata.items():
            if field not in options:
                raise RuntimeError(
                    "%s is not a valid option in the `[tool.ozi-build.metadata]` section, "
                    "got value: %s" % (field, value)
                )
            del options[field]
        for field, desc in options.items():
            if desc.get('required'):
                raise RuntimeError(
                    "%s is mandatory in the `[tool.ozi-build.metadata] section but was not found"
                    % field
                )
        pyc_whl_options = VALID_PYC_WHEEL_OPTIONS.copy()
        for field, value in self.__pyc_wheel.items():
            if field not in pyc_whl_options:
                raise RuntimeError(
                    "%s is not a valid option in the `[tool.ozi-build.pyc_wheel]` section, "
                    "got value: %s" % (field, value)
                )
            del pyc_whl_options[field]

    def get(self, key, default=None):
        return self.__metadata.get(key, default)

    def get_metadata(self):
        meta = {
            'name': self['module'],
            'version': self['version'],
        }
        pkg_info_file = check_pkg_info_file(self, meta)
        if pkg_info_file is not None:
            return pkg_info_file

        res = check_requires_python(
            self, auto_python_version(self, get_python_bin(self), meta)
        )
        res += get_optional_dependencies(self)
        res += get_simple_headers(self)
        res += get_license_headers(self)
        res += get_download_url_headers(self)
        res += get_requirements_headers(self)
        res += get_description_headers(self)

        return res






import sys
import unicodedata
from enum import Enum, auto
from typing import Set


class Category(Enum):
    DIGIT = auto()
    NOT_DIGIT = auto()
    WORD = auto()
    NOT_WORD = auto()
    SPACE = auto()
    NOT_SPACE = auto()

    @property
    def is_positive(self) -> bool:
        return not self.name.startswith("NOT_")

    def negate(self) -> "Category":
        if self.is_positive:
            return Category[f"NOT_{self.name}"]
        else:
            return Category[self.name[4:]]

    def example(self) -> str:
        return EXAMPLE_FOR_CAT[self]

    def contains(self, literal: int) -> bool:
        c = chr(literal)
        unicat = unicodedata.category(c)
        if self is Category.DIGIT:
            return unicat == "Nd"
        if self is Category.NOT_DIGIT:
            return unicat != "Nd"
        if self is Category.WORD:
            return (
                unicat[0] == "L" or unicat == "Nd" or literal == 0x5F
            )  # underscore is a word character
        if self is Category.NOT_WORD:
            return unicat[0] != "L" and unicat != "Nd" and literal != 0x5F
        if self is Category.SPACE:
            return unicat == "Zs" or c in (" ", "\n", "\t", "\r", "\f", "\v")
        if self is Category.NOT_SPACE:
            return unicat != "Zs" and c not in (" ", "\n", "\t", "\r", "\f", "\v")


CATS = {}


def list_category(category, full_unicode: bool = False):
    if (cached := CATS.get(category)) :
        yield from cached
    for data in range((sys.maxunicode + 1) if full_unicode else 256):
        c = chr(data)
        unicat = unicodedata.category(c)
        if category is Category.DIGIT:
            if unicat == "Nd":
                yield data
        elif category is Category.NOT_DIGIT:
            if unicat != "Nd":
                yield data
        elif category is Category.WORD:
            if unicat[0] == "L" or unicat == "Nd" or data == 0x5F:
                yield data
        elif category is Category.NOT_WORD:
            if unicat[0] != "L" and unicat != "Nd" and data != 0x5F:
                yield data
        elif category is Category.SPACE:
            if unicat == "Zs" or c in (" ", "\n", "\t", "\r", "\f", "\v"):
                yield data
        elif category is Category.NOT_SPACE:
            if unicat != "Zs" and c not in (" ", "\n", "\t", "\r", "\f", "\v"):
                yield data


def covers_any(categories: Set[Category]) -> bool:
    for c in categories:
        if c.is_positive and c.negate() in categories:
            return True
    return False


# CATS[sre_parse.CATEGORY_DIGIT] = list(list_category(sre_parse.CATEGORY_DIGIT))
# CATS[sre_parse.CATEGORY_SPACE] = list(list_category(sre_parse.CATEGORY_SPACE))
# CATS[sre_parse.CATEGORY_WORD] = list(list_category(sre_parse.CATEGORY_WORD))
EXAMPLE_FOR_CAT = {
    Category.DIGIT: "4",
    Category.NOT_DIGIT: "!",
    Category.WORD: "w",
    Category.NOT_WORD: "$",
    Category.SPACE: " ",
    Category.NOT_SPACE: ".",
}



import contextlib
import os
import subprocess
import sys
from fileinput import filename
import re

from .pep425tags import get_abbr_impl
from .pep425tags import get_abi_tag
from .pep425tags import get_impl_ver
from .pep425tags import get_platform_tag
from ._text import TextOutput
from ._redos import find
from ._sre import SreOpParser


class PotentialRedos(RuntimeError):
    ...


def handle_file(tomldata, filename: str, output: TextOutput):
    if isinstance(tomldata, (list, dict)):
        TomlWalker(filename, output).handle(tomldata)


class TomlWalker:
    def __init__(self, filename: str, output: TextOutput):
        self.filename = filename
        self.output = output

    def handle(self, elem):
        if isinstance(elem, str) and len(elem) > 5:
            try:
                parsed = SreOpParser().parse_sre(elem)
            except re.error:
                return  # We will have many strings which aren't actually regexes
            try:
                self.output.next()
                for redos in find(parsed):
                    if redos.starriness > 2:
                        self.output.record(
                            redos ,
                            elem,
                            filename=self.filename,
                        )
                        raise PotentialRedos(redos, elem, filename)
            except Exception as e:
                raise e
        elif isinstance(elem, list):
            for _elem in elem:
                self.handle(_elem)
        elif isinstance(elem, dict):
            for _elem in elem.values():
                self.handle(_elem)


def check_pyproject_regexes(file):
    output = TextOutput()
    handle_file(file, 'pyproject.toml', output)


PKG_INFO = """\
Metadata-Version: 2.2
Requires-Python: >={min_python}, <{max_python}
Name: {name}
Version: {version}
"""

PKG_INFO_CONFIG_REQUIRES_PYTHON = """\
Metadata-Version: 2.2
Requires-Python: {requires_python}
Name: {name}
Version: {version}
"""

PKG_INFO_NO_REQUIRES_PYTHON = """\
Metadata-Version: 2.2
Name: {name}
Version: {version}
"""

readme_ext_to_content_type = {
    '.rst': 'text/x-rst',
    '.md': 'text/markdown',
    '.txt': 'text/plain',
    '': 'text/plain',
}

GET_PYTHON_VERSION = 'import sys;print("{}.{}".format(*sys.version_info[:2]))'
GET_CHECK = """
from ozi_build import pep425tags
tag = pep425tags.get_abbr_impl() + pep425tags.get_impl_ver()
if tag != pep425tags.get_abi_tag():
    print("{0}-{1}".format(tag, pep425tags.get_abi_tag()))
else:
    print("{0}-none".format(tag))
"""

@contextlib.contextmanager
def cd(path):
    CWD = os.getcwd()

    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(CWD)


wheel_file_template = """\
Wheel-Version: 1.0
Generator: ozi_build
Root-Is-Purelib: {}
"""


def _write_wheel_file(f, supports_py2, is_pure):
    f.write(wheel_file_template.format(str(is_pure).lower()))
    if is_pure:
        if supports_py2:
            f.write("Tag: py2-none-any\n")
        f.write("Tag: py3-none-any\n")
    else:
        f.write(
            "Tag: {0}{1}-{2}-{3}\n".format(
                get_abbr_impl(),
                get_impl_ver(),
                get_abi_tag(),
                get_platform_tag(),
            )
        )


def install_files_path(installpath, target):
    while os.path.basename(installpath) != target:
        installpath = os.path.dirname(installpath)
    return installpath

def meson(*args, config_settings=None, builddir=''):
    try:
        return subprocess.check_output(['meson'] + list(args))
    except subprocess.CalledProcessError as e:
        stdout = ''
        stderr = ''
        if e.stdout:
            stdout = e.stdout.decode()
        if e.stderr:
            stderr = e.stderr.decode()
        print(
            "Could not run meson: %s\n%s" % (stdout, stderr), file=sys.stderr
        )
        try:
            fulllog = os.path.join(builddir, 'meson-logs', 'meson-log.txt')
            with open(fulllog) as f:
                print("Full log: %s" % f.read())
        except IOError:
            print("Could not open %s" % fulllog)
            pass
        raise e


def meson_configure(*args, config_settings=None):
    if 'MESON_ARGS' in os.environ:
        args = os.environ.get('MESON_ARGS').split(' ') + list(args)
        print("USING MESON_ARGS: %s" % args)
    args = list(args)
    args.append('-Dlibdir=lib')

    meson(*args, builddir=args[0], config_settings=config_settings)




import string
from dataclasses import dataclass
from typing import Optional, Set

from ._categories import Category, covers_any, list_category
from ._ranges import Range, lits_to_ranges


@dataclass(frozen=True)
class Character:
    literals: Optional[Set[int]] = None
    categories: Optional[Set[Category]] = None
    positive: bool = True

    @staticmethod
    def ANY() -> "Character":
        return Character()

    @staticmethod
    def LITERAL(literal: int) -> "Character":
        return Character({literal})

    @property
    def minimum_length(self) -> int:
        return 1

    @property
    def starriness(self) -> int:
        return 0

    def __hash__(self) -> int:
        return hash(
            (
                self.positive,
                tuple(sorted(self.literals)) if self.literals else None,
                tuple(sorted(self.categories)) if self.categories else None,
            )
        )

    def exact_character_class(self) -> "Character":
        return self

    def overall_character_class(self) -> "Character":
        return self

    def maximal_character_class(self) -> "Character":
        return self

    @property
    def is_any(self) -> bool:
        return self.literals is None and self.categories is None and self.positive

    @property
    def _is_positive_literal(self) -> bool:
        return self.positive and self.literals is not None and self.categories is None

    @property
    def _is_negative_literal(self) -> bool:
        return (
            not self.positive and self.literals is not None and self.categories is None
        )

    @property
    def _is_positive_category(self) -> bool:
        return self.positive and self.literals is None and self.categories is not None

    @property
    def _is_negative_category(self) -> bool:
        return (
            not self.positive and self.literals is None and self.categories is not None
        )

    def expand_categories(self) -> "Character":
        """
        This is the nuclear option where we expand the categories into literals.
        Can be huge in unicode.
        """
        if self.categories:
            lits: Set[int] = set(self.literals) if self.literals else set()
            for c in self.categories:
                lits.update(list_category(c))
            return Character(literals=lits, positive=self.positive)

        return self

    def __and__(self, other: "Optional[Character]") -> "Optional[Character]":
        if other is None:
            return None
        if self.is_any:
            return other
        if other.is_any:
            return self

        # [ab] & [bc] -> [c]
        if self._is_positive_literal and other._is_positive_literal:
            lits = self.literals & other.literals
            if not lits:
                return None
            return Character(literals=lits)
        if self._is_positive_category and other._is_positive_category:
            cats = self.categories & other.categories
            if not cats:
                return None
            return Character(categories=cats)
        # [^ab] & [^bc] -> [^abc]
        if self._is_negative_literal and other._is_negative_literal:
            return Character(literals=self.literals | other.literals, positive=False)
        if self._is_negative_category and other._is_negative_category:
            categories = self.categories | other.categories
            if covers_any(categories):  # [^\d] & [^\D] = nothing
                return None
            return Character(categories=categories, positive=False)
        # [ab] & [^bc] -> [a]
        if self._is_positive_literal and other._is_negative_literal:
            lits = self.literals - other.literals
            if not lits:
                return None
            return Character(literals=lits)
        if other._is_positive_literal and self._is_negative_literal:
            lits = other.literals - self.literals
            if not lits:
                return None
            return Character(literals=lits)

        # TODO: be less lazy and sort out the general case without expanding everything if possible
        return self.expand_categories() & other.expand_categories()

    def __rand__(self, other: "Optional[Character]") -> "Optional[Character]":
        return self & other

    def __or__(self, other: "Optional[Character]") -> "Optional[Character]":
        if other is None:
            return self
        if self.is_any or other.is_any:
            return Character.ANY()
        if self == other:
            return self
        if nor := (self.negate() & other.negate()):  # Slow, but logical
            return nor.negate()
        else:
            return Character.ANY()

    def __ror__(self, other: "Optional[Character]") -> "Optional[Character]":
        return self | other

    def __repr__(self) -> str:
        if self.is_any:
            return "."
        result = "["
        if not self.positive:
            result += "^"
        more = False
        if self.literals is not None:
            lits, ranges = lits_to_ranges(self.literals)
            result += ",".join(literal_repr(o) for o in lits)
            if lits and ranges:
                result += ","
            result += ",".join(range_repr(r) for r in ranges)
            more = True
        if self.categories is not None:
            if more:
                result += ";"
            result += ",".join(c.name for c in self.categories)
            more = True
        return result + "]"

    def example(self) -> str:
        for c in nice_characters():
            if self.matches(c):
                return chr(c)

        if self.positive:
            if self.literals:
                if len(self.literals) > 1:
                    # Try to avoid \n due to false positives with the . character and flags
                    return chr(next(o for o in self.literals if o != 0xA))
                return chr(next(iter(self.literals)))
            elif self.categories:
                return sorted(self.categories, key=lambda c: 0 if c.is_positive else 1)[
                    0
                ].example()

        raise NotImplementedError(self)

    def negate(self) -> "Optional[Character]":
        if self.is_any:
            return None
        return Character(
            literals=self.literals,
            categories=self.categories,
            positive=not self.positive,
        )

    def contains(self, subgroup: "Character") -> bool:
        if self.is_any:
            return True
        if subgroup.is_any:
            return False
        if subgroup == self:
            return True

        if self._is_positive_literal and subgroup._is_positive_literal:
            return not (subgroup.literals - self.literals)
        if self._is_positive_category and subgroup._is_positive_category:
            return not (subgroup.categories - self.categories)

        raise NotImplementedError  # Lazy, TODO: do full match

    def matches(self, literal: int) -> bool:
        if self.is_any:
            return True
        if self.literals is not None and literal in self.literals:
            return self.positive
        if self.categories:
            for cat in self.categories:
                if cat.contains(literal):
                    return self.positive
        return not self.positive


def nice_characters():
    for c in string.printable[:-5]:
        yield ord(c)


def literal_repr(literal: int) -> str:
    c = chr(literal)
    if c in string.digits or c in string.ascii_letters:
        return c
    elif c in string.punctuation:
        return f"{literal:02x}:{c}"
    return f"{literal:02x}"


def range_repr(r: Range) -> str:
    return "[{}-{}]".format(literal_repr(r.min_val), literal_repr(r.max_val))



from dataclasses import dataclass
from typing import Iterator, List, Set, Tuple


@dataclass(frozen=True)
class Range:
    min_val: int
    max_val: int


def lits_to_ranges(
    literals: Iterator[int],
) -> Tuple[Set[int], Set[Range]]:
    lits = set()
    ranges = set()
    buf: List[int] = []
    for lit in sorted(literals):
        if len(buf) and buf[-1] != lit - 1:
            # Discontinuity
            if len(buf) < 3:
                lits.update(buf)
            else:
                ranges.add(Range(buf[0], buf[-1]))
            buf = [lit]
        else:
            buf.append(lit)

    if len(buf) == 1:
        lits.add(buf[0])
    elif len(buf) > 1:
        ranges.add(Range(buf[0], buf[-1]))

    return lits, ranges
